import numpy as np
import pandas as pd
from mymodels.utils import accuracy_score

class Node():
    def __init__(self, feature_index=None, threshold=None, left=None, right=None, info_gain=None, value=None):
        ''' constructor ''' 
        
        # for decision node
        self.feature_index = feature_index
        self.threshold = threshold
        self.left = left
        self.right = right
        self.info_gain = info_gain
        
        # for leaf node
        self.value = value
        
class DecisionTree:
    def __init__(self, min_samples_split=2, max_depth=3, criterion="gini",n_features=None):
        """Constructor to initialize the tree parameters."""
        self.root = None
        self.min_samples_split = min_samples_split
        self.max_depth = max_depth
        self.criterion = criterion # Criterion can be 'gini' or 'entropy'
        self.n_features = n_features
    
    def build_tree(self, dataset, curr_depth=0):
        """Recursive function to build the tree."""
        X, Y = dataset[:, :-1], dataset[:, -1]
        num_samples, num_features = np.shape(X)
        
        if num_samples >= self.min_samples_split and curr_depth <= self.max_depth:
            # Find the best split
            best_split = self.get_best_split(dataset, num_samples, num_features)
            
            # Check if the information gain is positive
            if best_split["info_gain"] > 0:
                # Build the left and right subtrees recursively
                left_subtree = self.build_tree(best_split["dataset_left"], curr_depth + 1)
                right_subtree = self.build_tree(best_split["dataset_right"], curr_depth + 1)
                
                # Return a decision node
                return Node(
                    feature_index=best_split["feature_index"],
                    threshold=best_split["threshold"],
                    left=left_subtree,
                    right=right_subtree,
                    info_gain=best_split["info_gain"]
                )
        
        # Compute the leaf node value
        leaf_value = self.calculate_leaf_value(Y)
        return Node(value=leaf_value)
    
    def get_best_split(self, dataset, num_samples, num_features):
        """Find the best split for the data."""
        best_split = {"info_gain": 0}
        max_info_gain = -1
        
        # Iterate over all features
        for feature_index in range(num_features):
            feature_values = dataset[:, feature_index]
            possible_thresholds = np.unique(feature_values)
            
            # Iterate over all unique threshold values for the feature
            for threshold in possible_thresholds:
                dataset_left, dataset_right = self.split(dataset, feature_index, threshold)
                
                if len(dataset_left) > 0 and len(dataset_right) > 0:
                    y = dataset[:, -1]
                    left_y = dataset_left[:, -1]
                    right_y = dataset_right[:, -1]
                    
                    # Calculate information gain based on the criterion
                    curr_info_gain = self.information_gain(y, left_y, right_y)
                    
                    # Update the best split if a better one is found
                    if curr_info_gain > max_info_gain:
                        best_split = {
                            "feature_index": feature_index,
                            "threshold": threshold,
                            "dataset_left": dataset_left,
                            "dataset_right": dataset_right,
                            "info_gain": curr_info_gain
                        }
                        max_info_gain = curr_info_gain
        
        return best_split
    
    def split(self, dataset, feature_index, threshold):
        """Split the dataset into two based on a threshold."""
        dataset_left = np.array([row for row in dataset if row[feature_index] <= threshold])
        dataset_right = np.array([row for row in dataset if row[feature_index] > threshold])
        return dataset_left, dataset_right
    
    def information_gain(self, parent, l_child, r_child):
        """Compute the information gain."""
        weight_l = len(l_child) / len(parent)
        weight_r = len(r_child) / len(parent)
        
        if self.criterion == "gini":
            gain = self.gini_index(parent) - (weight_l * self.gini_index(l_child) + weight_r * self.gini_index(r_child))
        else:  # Use entropy
            gain = self.entropy(parent) - (weight_l * self.entropy(l_child) + weight_r * self.entropy(r_child))
        
        return gain
    
    def entropy(self, y):
        """Compute entropy."""
        class_labels = np.unique(y)
        entropy = 0
        for cls in class_labels:
            p_cls = len(y[y == cls]) / len(y)
            entropy += -p_cls * np.log2(p_cls)
        return entropy
    
    def gini_index(self, y):
        """Compute Gini index."""
        class_labels = np.unique(y)
        gini = 0
        for cls in class_labels:
            p_cls = len(y[y == cls]) / len(y)
            gini += p_cls**2
        return 1 - gini
    
    def calculate_leaf_value(self, Y):
        """Compute the value of a leaf node."""
        Y = list(Y)
        return max(Y, key=Y.count)
    
    def fit(self, X, Y):
        """Train the decision tree."""
        # Ensure Y is 2D for concatenation
        if len(Y.shape) == 1:
            Y = Y.reshape(-1, 1)
        dataset = np.concatenate((X, Y), axis=1)
        self.root = self.build_tree(dataset)

    
    def predict(self, X):
        """Predict for a dataset."""
        predictions = [self.make_prediction(x, self.root) for x in X]
        return predictions
    
    def make_prediction(self, x, tree):
        """Predict for a single data point."""
        if tree.value is not None:
            return tree.value
        
        feature_val = x[tree.feature_index]
        if feature_val <= tree.threshold:
            return self.make_prediction(x, tree.left)
        else:
            return self.make_prediction(x, tree.right)        
    def score(self, X, y):
        predictions = self.predict(X)
        return accuracy_score(y, predictions)
    
